import torch.nn as nn
from .gated_units import GatedLinear


class PermuteTF(nn.Module):
    def forward(self, x):
        return x.transpose(1, 2).contiguous()


class GatedLinearInput(GatedLinear):
    def __init__(self, in_dim, out_dim, permute_tf=True):
        super(GatedLinearInput, self).__init__(in_dim, out_dim)
        self.post = PermuteTF() if permute_tf else nn.Identity()

    def forward(self, x):
        x = super(GatedLinearInput, self).forward(x)
        return self.post(x)


class AbsLinearOutput(nn.Module):
    def __init__(self, in_dim, out_dim, permute_tf=True, **kwargs):
        super(AbsLinearOutput, self).__init__()
        self.fc = nn.Sequential(
            PermuteTF() if permute_tf else nn.Identity(),
            nn.Linear(in_dim, out_dim, **kwargs))

    def forward(self, x):
        return self.fc(x).abs()



from .base import FreqNetModel, FreqOptim, FreqData
from .freaksnet import FreaksNet
from .freq_layer import FreqLayer
from .freqnet import FreqNet
from .frexpert_mixture import FrexpertMixture
from .layer_wise_loss import LayerWiseLossFreqNet
from .hierarchical_kernel import HKFreqNet
from .freqnet_db import freqnet_db



import h5py
import numpy as np
import pandas as pd
from multiprocessing import cpu_count, Pool
from typing import Iterable
import os
import warnings

from .api import Database
from .metadata import Metadata
from .transforms import default_extract_func

warnings.filterwarnings("ignore", message="PySoundFile failed.")
warnings.filterwarnings("ignore", message="PerformanceWarning")
warnings.filterwarnings("ignore", message="Creating an ndarray from ragged nested sequences")


class AudioFileWalker:

    AUDIO_EXTENSIONS = ("wav", "aif", "aiff", "mp3", "m4a", "mp4")

    def __init__(self, roots=None, files=None):
        """
        recursively find audio files from `roots` and/or collect audio_files passed in `files`
        @param roots: a single path (string, os.Path) or an Iterable of paths
                      from which to collect audio files recursively
        @param files: a single path (string, os.Path) or an Iterable of paths
        N.B. : any file whose extension isn't in AudioFileWalker.AUDIO_EXTENSIONS will be ignored,
        regardless whether it was found recursively or passed through the `files` argument.

        AudioFileWalker implements `__iter__`, hence files can be retrieved in different ways.
        For instance so:
        ```
        files = list(AudioFileWalker(roots=some_roots, files=some_files))
        ```
        """
        generators = []

        if roots is not None and isinstance(roots, Iterable):
            if isinstance(roots, str):
                if not os.path.exists(roots):
                    raise FileNotFoundError("%s does not exist." % roots)
                generators += [AudioFileWalker.walk_root(roots)]
            else:
                for r in roots:
                    if not os.path.exists(r):
                        raise FileNotFoundError("%s does not exist." % r)
                generators += [AudioFileWalker.walk_root(root) for root in roots]

        if files is not None and isinstance(files, Iterable):
            if isinstance(files, str):
                if not os.path.exists(files):
                    raise FileNotFoundError("%s does not exist." % files)
                generators += [(f for f in [files] if AudioFileWalker.is_audio_file(files))]
            else:
                for f in files:
                    if not os.path.exists(f):
                        raise FileNotFoundError("%s does not exist." % f)
                generators += [(f for f in files if AudioFileWalker.is_audio_file(f))]

        self.generators = generators

    def __iter__(self):
        for generator in self.generators:
            for file in generator:
                yield file

    @staticmethod
    def walk_root(root):
        for directory, _, files in os.walk(root):
            for audio_file in filter(AudioFileWalker.is_audio_file, files):
                yield os.path.join(directory, audio_file)

    @staticmethod
    def is_audio_file(filename):
        # filter out hidden files (isn't cross-platform, but, it's a start!...)
        if filename.startswith("."):
            return False
        return os.path.splitext(filename)[-1].strip(".") in AudioFileWalker.AUDIO_EXTENSIONS


def sizeof_fmt(num, suffix='b'):
    """
    straight from https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    """
    for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def _empty_info(features_names):
    tuples = [("directory", ""), ("name", ""),
              *[t for feat in features_names for t in [(feat, "dtype"), (feat, "shape"), (feat, "size")]
                if feat != "metadata"]
              ]
    idx = pd.MultiIndex.from_tuples(tuples)
    return pd.DataFrame([], columns=idx)


def split_path(path):
    prefix, file_name = os.path.split(path)
    return prefix, file_name


# Core function

def file_to_db(abs_path, extract_func=default_extract_func, mode="w"):
    """
    if mode == "r+" this will either:
        - raise an Exception if the feature already exists
        - concatenate data along the "feature_axis", assuming that each feature correspond to the same file
          or file collections.
          If you want to concatenate dbs along the "file_axis" consider using `concatenate_dbs(..)`
    @param abs_path:
    @param extract_func:
    @param mode:
    @return:
    """
    print("making db for %s" % abs_path)
    tmp_db = os.path.splitext(abs_path)[0] + ".h5"
    rv = extract_func(abs_path)
    if "metadata" not in rv:
        raise ValueError("Expected `extract_func` to return a ('metadata', Metadata) item. Found none")
    info = _empty_info(rv.keys())
    info.loc[0, [("directory", ""), ("name", "")]] = split_path(abs_path)
    f = h5py.File(tmp_db, mode)
    for name, (attrs, data) in rv.items():
        if issubclass(type(data), np.ndarray):
            ds = f.create_dataset(name=name, shape=data.shape, data=data)
            ds.attrs.update(attrs)
            info.at[0, [(name, "dtype"), (name, "shape"), (name, "size")]] = tuple([ds.dtype, ds.shape, sizeof_fmt(data.nbytes)])
        elif issubclass(type(data), pd.DataFrame):
            f.close()
            pd.DataFrame(data).to_hdf(tmp_db, name, "r+")
            f = h5py.File(tmp_db, "r+")
    f.flush()
    f.close()
    if "info" in f.keys():
        prior = pd.read_hdf(tmp_db, "info", "r")
        info = pd.concat((prior, info.iloc[:, 2:]), axis=1)
    info.to_hdf(tmp_db, "info", "r+")
    return tmp_db


# Multiprocessing routine

def make_db_for_each_file(file_walker,
                          extract_func=default_extract_func,
                          n_cores=cpu_count()):
    args = [(file, extract_func) for file in file_walker]
    with Pool(n_cores) as p:
        tmp_dbs = p.starmap(file_to_db, args)
    return tmp_dbs


# Aggregating sub-tasks

def collect_infos(tmp_dbs):
    infos = []
    for db in tmp_dbs:
        infos += [Database(db).info]
    return pd.concat(infos, ignore_index=True)


def collect_metadatas(tmp_dbs):
    metadatas = []
    offset = 0
    for db in tmp_dbs:
        meta = Database(db).metadata
        meta.loc[:, ("start", "stop")] = meta.loc[:, ("start", "stop")].values + offset
        meta.loc[:, "name"] = os.path.splitext(db)[0]
        metadatas += [meta]
        offset = meta.last_stop
    return pd.DataFrame(pd.concat(metadatas, ignore_index=True))


def zip_prev_next(iterable):
    return zip(iterable[:-1], iterable[1:])


def ds_definitions_from_infos(infos):
    tb = infos.iloc[:, 2:].T
    paths = [os.path.join(*parts) for parts in infos.iloc[:, :2].values]
    # change the paths' extensions
    paths = [os.path.splitext(path)[0] + ".h5" for path in paths]
    features = set(tb.index.get_level_values(0))
    ds_definitions = {}
    for f in features:
        dtype = tb.loc[(f, "dtype"), :].unique().item()
        shapes = tb.loc[(f, "shape"), :].values
        dims = shapes[0][1:]
        assert all(shp[1:] == dims for shp in
                   shapes[1:]), "all features should have the same dimensions but for the first axis"
        layout = Metadata.from_duration([s[0] for s in shapes])
        ds_shape = (layout.last_stop, *dims)
        layout.index = paths
        ds_definitions[f] = {"shape": ds_shape, "dtype": dtype, "layout": layout}
    return ds_definitions


def create_datasets_from_defs(target, defs, mode="w"):
    f = h5py.File(target, mode)
    for name, params in defs.items():
        f.create_dataset(name, shape=params["shape"], dtype=params["dtype"],
                         chunks=True, maxshape=(None, *params["shape"][1:]))
        layout = params["layout"]
        layout.reset_index(drop=False, inplace=True)
        layout = layout.rename(columns={"index": "name"})
        f.flush()
        f.close()
        pd.DataFrame(layout).to_hdf(target, "layouts/" + name, "r+", format="table")
        f = h5py.File(target, "r+")
    f.flush()
    f.close()
    return


def make_integration_args(target):
    args = []
    with h5py.File(target, "r") as f:
        for feature in f["layouts"].keys():
            df = Metadata(pd.read_hdf(target, "layouts/" + feature))
            args += [(target, source, feature, indices) for source, indices in
                     zip(df.name, df.slices(time_axis=0))]
    return args


def integrate(target, source, key, indices):
    with h5py.File(source, "r") as src:
        data = src[key][()]
        attrs = {k: v for k, v in src[key].attrs.items()}
    with h5py.File(target, "r+") as trgt:
        trgt[key][indices] = data
        trgt[key].attrs.update(attrs)
    return


# Aggregating function and main client

def aggregate_dbs(target, dbs, mode="w", remove_sources=False):
    infos = collect_infos(dbs)
    metadata = collect_metadatas(dbs)
    definitions = ds_definitions_from_infos(infos)
    create_datasets_from_defs(target, definitions, mode)
    args = make_integration_args(target)
    for arg in args: integrate(*arg)
    if remove_sources:
        for src in dbs: os.remove(src)
    infos = infos.astype(object)
    infos.to_hdf(target, "info", "r+")
    metadata.to_hdf(target, "metadata", "r+")


def make_root_db(db_name, roots='./', files=None, extract_func=default_extract_func,
                 n_cores=cpu_count(), remove_sources=True):
    walker = AudioFileWalker(roots, files)
    dbs = make_db_for_each_file(walker, extract_func, n_cores)
    aggregate_dbs(db_name, dbs, "w", remove_sources)
    return Database(db_name)



from pytorch_lightning.utilities.cloud_io import load as pl_load
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from time import time, gmtime
from typing import Optional, Union, Callable, Dict, IO
import torch
import os
import warnings
import soundfile as sf

try:
    from neptune.experiments import Experiment as NeptuneExperiment
    from neptunecontrib.api.audio import log_audio
except ImportError:
    NeptuneExperiment = type(None)
    warnings.warn("It seems neptune and/or neptunecontrib are not installed. "
                  "You won't be able to log checkpoints and data to neptune.")

from .. import __version__ as version
from ..data.transforms import SR


class LoggingHooks:

    @property
    def neptune_experiment(self):
        """ short hand to access own NeptuneExperiment"""
        exps = [exp for exp in self.logger.experiment
               if isinstance(exp, NeptuneExperiment)]
        if any(exps):
            return exps[0]
        return None

    def on_epoch_start(self):
        # convenience to have printing of the loss at the end of the epoch
        self._ep_metrics = {}
        self._batch_count = {}

    def log_output(self, out):
        for metric, val in out.items():
            if metric not in self._ep_metrics:
                self._ep_metrics.setdefault(metric, val.detach())
                self._batch_count[metric] = 0
            else:
                self._ep_metrics[metric] += val.detach()
            self._batch_count[metric] += 1
        return out

    def training_step_end(self, out):
        return self.log_output(out)

    def validation_step_end(self, out):
        return self.log_output(out)

    def on_train_epoch_end(self, *args):
        to_print = "Epoch %i " % self.current_epoch
        to_log = {}
        for k, v in self._ep_metrics.items():
            to_print += "- %s : %.4f " % (k, v / self._batch_count[k])
            to_log[k] = (v / self._batch_count[k]).item()
        self.print(to_print)
        if getattr(self, "logger", None) is not None:
            self.logger.log_metrics(to_log)

    def on_fit_start(self):
        self._ep_time = time()

    def on_fit_end(self):
        duration = time() - self._ep_time
        total_time = gmtime(duration)
        self.print("Training finished after "
                   "{0} days {1} hours {2} mins {3} seconds".format(total_time[2] - 1, total_time[3],
                                                                    total_time[4], total_time[5]))
        if getattr(self, "logger", None) is not None:
            # accumulate if we resumed
            duration = self.hparams.get("training_time_sec", 0) + duration
            self.hparams.update({"training_time_sec": duration})
            self.logger.log_hyperparams(self.hparams)

    def log_audio(self, filename, audio_tensor, sample_rate=SR, experiments=None):
        # TensorBoards write their own "Event" file which is quite cumbersome to extract afterwards...
        # NeptuneLoggers need an audio file written on disk and store it afterwards as html on the server...
        # Just to be sure and even if it is then in 3 places : we add the audio in root_dir/audios of the model!

        # figure out where we'll save the file :
        if self.trainer is not None:
            root = os.path.join(self.trainer.default_root_dir, "audios")
        elif self._loaded_checkpoint is not None:
            root_dir = os.path.split(os.path.dirname(self._loaded_checkpoint))[0]
            root = os.path.join(root_dir, "audios")
        else:
            if os.path.dirname(filename):
                root = os.path.dirname(filename)
            else:
                root = "./"
        if not os.path.exists(root):
            os.makedirs(root, exist_ok=True)
        path = os.path.join(root, filename)
        if ".wav" != os.path.splitext(path)[-1]:
            path = path + ".wav"
        audio_tensor = audio_tensor.squeeze().detach().cpu().numpy()
        sf.write(path, audio_tensor, sample_rate, 'PCM_24')
        exps = [] if experiments is None else experiments
        exps += self.logger.experiment if getattr(self, "logger", False) else []
        for exp in exps:
            # Neptune and TestTube experiments have different APIs...
            if getattr(exp, "add_audio", False):
                exp.add_audio(filename + ".wav", audio_tensor, sample_rate=sample_rate)
                exp.save()
                print("Updated TensorBoard with", filename)
            elif isinstance(exp, NeptuneExperiment):
                log_audio(path, os.path.split(path)[-1], exp)
                print("Updated neptune experiment", exp.id, "with", filename)
        return 1


def _check_version(other_v):
    if other_v.split(".")[0] != version.split(".")[0]:
        v = str(other_v)
        warnings.warn(("You are loading a checkpoint made by a different version of mimikit (%s) as the one" % v) +
                      (" imported in this runtime (%s). If you encounter errors " % version) +
                      (", try to install the right version with `pip install mimikit==%s" % v))


class MMKHooks:

    _loaded_checkpoint = None

    @classmethod
    def load_from_checkpoint(
            cls,
            checkpoint_path: Union[str, IO],
            map_location: Optional[Union[Dict[str, str], str, torch.device, int, Callable]] = None,
            strict: bool = True,
            **kwargs):
        """
        This is NOT used when restoring a model from the trainer, but still practical in other cases!
        """
        if map_location is not None:
            checkpoint = pl_load(checkpoint_path, map_location=map_location)
        else:
            checkpoint = pl_load(checkpoint_path, map_location=lambda storage, loc: storage)

        if checkpoint.get("version", None) is not None:
            _check_version(checkpoint["version"])
        model = cls._load_model_state(checkpoint, strict=strict, **kwargs)
        model._loaded_checkpoint = checkpoint_path
        return model

    def on_save_checkpoint(self, checkpoint):
        ckpt_callbacks = [cb for cb in self.trainer.callbacks
                          if issubclass(type(cb), ModelCheckpoint)]
        if any(ckpt_callbacks):
            path = ckpt_callbacks[0].dirpath
        else:
            path = self.trainer.default_root_dir

        # overwrite previous optim state
        torch.save({"optimizer_states": checkpoint["optimizer_states"],
                    "lr_schedulers": checkpoint["lr_schedulers"]
                    }, os.path.join(path, "last_optim_state.ckpt"))

        # pop optim items from checkpoint
        checkpoint.pop("optimizer_states")
        checkpoint.pop("lr_schedulers")
        # this seem to fix ValueError caused by lr_scheduler when resuming:
        checkpoint["global_step"] -= 2
        checkpoint["mmk_version"] = version
        return checkpoint

    def on_load_checkpoint(self, checkpoint):
        if checkpoint.get("version", None) is not None:
            _check_version(checkpoint["version"])
        # we restore the training state only if the model has a trainer...
        if getattr(self, "trainer", None) is not None:
            # ... and the trainer got a ckpt_path to resume from
            if self.trainer.resume_from_checkpoint is not None:
                ckpt_path = self.trainer.resume_from_checkpoint
                version_dir = os.path.abspath(os.path.dirname(ckpt_path))

                # make sure we'll find the optim-state or let the user know
                if "last_optim_state.ckpt" not in os.listdir(version_dir):
                    raise FileNotFoundError("Expected to find 'last_optim_state.ckpt' in the parent directory "
                                            "of the model's checkpoint's directory but no such file were found.")
                # update the checkpoint
                optim_state = torch.load(os.path.join(version_dir, "last_optim_state.ckpt"))
                checkpoint["optimizer_states"] = optim_state["optimizer_states"]
                checkpoint["lr_schedulers"] = optim_state["lr_schedulers"]

        return checkpoint

    def upload_to_neptune(self, root_dir=None, experiment=None, artifacts=("states", "logs", "audios")):
        if root_dir is None:
            if getattr(self, "trainer", None) is None:
                raise ValueError("expected to either have a 'trainer' attribute or `root_dir` to be a"
                                 " valid path to a model's root directory. Both were None.")
            else:
                root_dir = self.trainer.default_root_dir
        if experiment is None:
            if not any(isinstance(exp, NeptuneExperiment) for exp in self.logger.experiment):
                raise ValueError("`experiment` is None and this model isn't bound to any NeptuneExperiment...")
            experiment = [exp for exp in self.logger.experiment if isinstance(exp, NeptuneExperiment)][0]
        # log everything!
        for directory in os.listdir(root_dir):
            if directory in artifacts:
                artifact = os.path.join(root_dir, directory)
                experiment.log_artifact(artifact, directory)
                print("successfully uploaded", artifact, "to", experiment.id)
        return 1





import torch
from torch import nn as nn


class LearnablePad1d(nn.Module):
    def __init__(self, input_dim, amount, learn_padding=False):
        super(LearnablePad1d, self).__init__()
        self.pad = None
        self.side = torch.sign(torch.tensor(amount)).item()
        amount = abs(amount)
        if amount > 0:
            if learn_padding:
                self.pad = nn.Parameter(torch.zeros(input_dim, amount), requires_grad=True)
            else:
                self.pad = torch.zeros(input_dim, amount)

    def forward(self, x):
        if self.pad is None:
            return x
        self.pad = self.pad.to(x)
        return torch.cat((torch.stack([self.pad] * x.size(0)), x)[::self.side], dim=-1)


    def receptive_field(self):
        shifts = []
        for i, layer in enumerate(self.layers):
            if shifts[-1] > layer.shift():
                shifts += [shifts[-1] + layer.shift()]
            else:
                shifts += [layer.shift()]
        return sum(shifts)


import torch
import torch.nn as nn

from .freqnet import FreqNet
from .modules import mean_L1_prop


class HKFreqNet(FreqNet):
    """
    Add a hierarchical quality to the conditioning sequence by combining FreqLayers in a
    final Conv2d making depth in the network proportional to the distance to the predicted time-step.

    N.B : Could be fun to try the other way around : the closer to now -> the deeper in the net.
    """

    def __init__(self,
                 loss_fn=mean_L1_prop,
                 model_dim=512,
                 groups=1,
                 n_layers=2,
                 strict=False,
                 accum_outputs=0,
                 concat_outputs=0,
                 pad_input=0,
                 learn_padding=False,
                 with_skip_conv=True,
                 with_residual_conv=True,
                 **data_optim_kwargs):
        super(HKFreqNet, self).__init__(loss_fn, model_dim, groups, (n_layers,), strict,
                                        accum_outputs, concat_outputs, pad_input, learn_padding,
                                        with_skip_conv, with_residual_conv,
                                        **data_optim_kwargs)
        # the 4th dimension in the final conv2d is 2 encoder outputs + n_conv_layers
        self.conv_f = nn.Conv2d(model_dim, model_dim, (1, n_layers + 2), groups=groups)
        self.conv_g = nn.Conv2d(model_dim, model_dim, (1, n_layers + 2), groups=groups)

    def forward(self, x):
        """

        """
        x = self.inpt(x)

        # the conv2d layer adds a power of 2 to the n standard layers before it
        n_outs = 1 + x.size(-1) - self.receptive_field() * 2

        # stack the 2 encoder outputs
        hk_input = torch.stack((x[:, :, -n_outs - 1:-1], x[:, :, -n_outs:]), dim=-1)
        skip = None

        for layer in self.layers:
            x, skip = layer(x, skip)

            # we only keep n_outs outputs finishing at the -`layer.shift()` step
            to_stack = slice(-n_outs - layer.shift(), -layer.shift())
            hk_input = torch.cat((hk_input, skip[:, :, to_stack].unsqueeze(-1)), dim=-1)

        # Do the Gated2dConv on the gathered pieces of layers
        f = nn.Tanh()(self.conv_f(hk_input))
        g = nn.Sigmoid()(self.conv_g(hk_input))

        y = self.outpt((f * g).squeeze(-1))
        return y

# TODO WIP : shifts and output_lengths !!!

    def all_rel_shifts(self):
        """sequence of shifts from one layer to the next"""
        stack = super(HKFreqNet, self).all_rel_shifts()
        return (*stack, 2 * stack[-1])

    def all_output_lengths(self, input_length):
        out_length = input_length
        lengths = []
        for layer in self.layers:
            out_length = layer.output_length(out_length)
            lengths += [out_length]
        # add the last layer
        lengths += [out_length - (len(self.layers) + 2)]
        return tuple(lengths)



import os
import torch
from pytorch_lightning.loggers import NeptuneLogger
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
from .loggers import MMKDefaultLogger
from .checkpoint import MMKCheckpoint
from .callbacks import EpochProgressBarCallback
from pytorch_lightning.trainer import Trainer
import warnings


def get_trainer(model=None,
                root_dir=None,
                version=None,
                resume_from_checkpoint=None,
                epochs=None,  # add MMKCheckpoint if not None
                neptune_api_token=None,  # add NeptuneLogger if not None
                neptune_project=None,
                neptune_exp_id=None,
                **kwargs):

    # Figure out the root_dir
    if root_dir is None:
        if resume_from_checkpoint is not None:
            # keep the same root_dir as the checkpoint
            ckpt_dir = os.path.abspath(os.path.dirname(resume_from_checkpoint))
            root_dir = os.path.dirname(ckpt_dir)
            kwargs["resume_from_checkpoint"] = resume_from_checkpoint
        else:
            root_dir = "./"
    else:
        if resume_from_checkpoint is not None:
            kwargs.setdefault("resume_from_checkpoint", resume_from_checkpoint)

    # Figure out the version
    if version is not None:
        if not isinstance(version, int):
            raise TypeError("Expected `version` to be of type int. Got %s" % str(type(version)))
        if version == -1:
            versions = [int(v.split("_")[1]) for v in os.listdir(root_dir) if "version" in v]
            next_version = 1 + (max(versions) if versions else -1)
        else:
            next_version = version  # this overwrites any existing files!
        default_root_dir = os.path.join(root_dir, "version_" + str(next_version))
    else:
        next_version = 0
        default_root_dir = root_dir

    # Figure out loggers
    if (neptune_api_token is None) != (neptune_project is None):
        raise ValueError("Expected `neptune_project` and `neptune_api_token` to both be either None or not None")

    user_logger = kwargs.get("logger", None)
    loggers = []

    if neptune_api_token is not None:
        if model is None:
            raise ValueError("Expected `model` not to be None in order to create a neptune.Experiment")
        loggers.append(NeptuneLogger(neptune_api_token, neptune_project, params=model.hparams,
                                     experiment_name=default_root_dir, experiment_id=neptune_exp_id))
    if user_logger is None:
        loggers.append(MMKDefaultLogger(default_root_dir, next_version))
    elif user_logger:
        loggers.append(user_logger)
    else:  # it is falsy and the user DOESN'T want logging
        loggers = False
        if neptune_api_token is not None:
            warnings.warn("You provided arguments to instantiate a NeptuneLogger but set `logger=False` in the kwargs. "
                          "The latter resulting in turning any logging off, your experiment won't be logged to neptune.")

    kwargs["logger"] = loggers

    # Figure out checkpoints
    # if the user specifies one, we don't add ours
    if not any(issubclass(type(cb), ModelCheckpoint) for cb in kwargs.get("callbacks", [])):
        if epochs is not None:
            kwargs.setdefault("callbacks", []).append(MMKCheckpoint(dirpath=default_root_dir, epochs=epochs))
        else:
            warnings.warn("You neither specified `epochs` to enable default checkpoint_callback "
                          "nor passed a `checkpoint_callback` argument as kwarg. Saving of your model won't be managed"
                          " by mimikit nor by pytorch_lightning's Trainer.")
    # add epoch progress bar
    kwargs.setdefault("callbacks", []).append(EpochProgressBarCallback())
    kwargs.setdefault("progress_bar_refresh_rate", 5)
    kwargs.setdefault("process_position", 1)
    kwargs.setdefault("num_sanity_val_steps", 0)  # this is 2 by default and messes up the steps count...
    kwargs.setdefault("gpus", torch.cuda.device_count() if torch.cuda.is_available() else 0)
    print("Checkpoints and logs will be saved in", os.path.abspath(default_root_dir))
    return Trainer(default_root_dir=default_root_dir, **kwargs)



import torch.nn as nn
import torch.nn.functional as F
from functools import partial
from multiprocessing import cpu_count
import os

from . import FreqNet
from .modules.io import PermuteTF
from ..data import Database, make_root_db, upload_database
from ..data.transforms import file_to_qx


def wavenet_db(target,
               roots=None,
               files=None,
               mu=255,
               sample_rate=22050,
               neptune_project=None):
    transform = partial(file_to_qx,
                        mu=mu,
                        sr=sample_rate)
    if roots is None and files is None:
        roots = "./"
    make_root_db(target, roots=roots, files=files, extract_func=transform, n_cores=cpu_count() // 2)

    if neptune_project is not None:
        token = os.environ["NEPTUNE_API_TOKEN"]
        db = Database(target)
        print("uploading database to neptune...")
        upload_database(db, token, neptune_project, target)
    print("done!")
    return Database(target)


def wavenet_loss_fn(output, target):
    return F.nll_loss(output.view(-1, output.size(-1)), target.view(-1))


class WaveNet(FreqNet):

    def __init__(self,
                 input_dim=256,
                 model_dim=256,
                 groups=1,
                 n_layers=(1,),
                 with_skip_conv=True,
                 with_residual_conv=True,
                 **data_optim_kwargs
                 ):
        super(WaveNet, self).__init__(
            loss_fn=wavenet_loss_fn,
            model_dim=model_dim,
            groups=groups,
            n_layers=n_layers,
            strict=False,
            accum_outputs="right",
            concat_outputs=None,
            pad_input="left",
            learn_padding=False,
            with_skip_conv=with_skip_conv,
            with_residual_conv=with_residual_conv,
            **data_optim_kwargs)

        self.input_dim = input_dim

        self.inpt = nn.Sequential(
            nn.Embedding(self.input_dim, self.model_dim), PermuteTF())

        self.outpt = nn.Sequential(
            nn.ReLU(inplace=True),
            nn.Conv1d(self.model_dim, self.model_dim, 1),
            nn.ReLU(inplace=True),
            nn.Conv1d(self.model_dim, self.input_dim, 1),
            PermuteTF(),
            nn.LogSoftmax(dim=1)
        )

        self.save_hyperparameters()

    def forward(self, x):
        x = self.inpt(x.squeeze())
        skips = None
        for layer in self.layers:
            x, skips = layer(x, skips)
        x = self.outpt(skips)
        return x



import torch.nn as nn

from .freq_layer import FreqLayer
from .freqnet import FreqNet
from .modules import mean_L1_prop, GatedLinearInput, AbsLinearOutput


class FreaksNet(FreqNet):
    """
    Simple adaptation of FreqNet to have varying parameters for each layer.
    to each block of layer corresponds a tuple of values for the params listed in FeqNet.LAYER_PARAMS
    """

    def __init__(self,
                 loss_fn=mean_L1_prop,
                 input_dim=1025,
                 model_dim=512,
                 groups=((1, 1, 1),),
                 n_layers=(3,),
                 strict=((False, True, False),),
                 accum_outputs=((-1, 1, -1),),
                 concat_outputs=((1, -1, 1),),
                 pad_input=((0, 0, 0),),
                 learn_padding=((False, False, False),),
                 with_skip_conv=((False, True, False,), ),
                 with_residual_conv=((True, True, False), ),
                 **data_optim_kwargs):
        super(FreaksNet, self).__init__(**data_optim_kwargs)
        self._loss_fn = loss_fn
        self.input_dim = input_dim
        self.model_dim = model_dim
        self.groups = groups
        self.n_layers = n_layers
        self.strict = strict
        self.accum_outputs = accum_outputs
        self.concat_outputs = concat_outputs
        self.pad_input = pad_input
        self.learn_padding = learn_padding
        self.with_skip_conv = with_skip_conv
        self.with_residual_conv = with_residual_conv

        # Input Encoder
        self.inpt = GatedLinearInput(self.input_dim, self.model_dim)

        # Auto-regressive Part
        layer_kwargs = {attr: getattr(self, attr) for attr in self.LAYER_KWARGS}
        # for simplicity we keep all the layers in a flat list
        self.layers = nn.ModuleList([
            FreqLayer(layer_index=i, input_dim=model_dim, layer_dim=model_dim,
                      **{k: v[j][i] for k, v in layer_kwargs.items()})
            for j, n_layers in enumerate(self.n_layers) for i in range(n_layers)
        ])

        # Output Decoder
        self.outpt = AbsLinearOutput(self.model_dim, self.input_dim)

        self.save_hyperparameters()



import h5py
import numpy as np
import pandas as pd
import os
from datetime import datetime

from neptune import Session
from torch.utils.data.dataset import Subset

from .metadata import Metadata
from ..data.data_object import DataObject


class FeatureProxy(object):

    def __init__(self, h5_file, ds_name):
        self.h5_file = h5_file
        self.name = ds_name
        with h5py.File(h5_file, "r") as f:
            ds = f[self.name]
            self.N = ds.shape[0]
            self.shape = ds.shape
            self.attrs = {k: v for k, v in ds.attrs.items()}

    def __len__(self):
        return self.N

    def __getitem__(self, item):
        with h5py.File(self.h5_file, "r") as f:
            rv = f[self.name][item]
        return rv

    def get(self, metadata):
        t_axis = self.attrs.get("time_axis", 0)
        slices = metadata.slices(t_axis)
        return np.concatenate(tuple(self[slice_i] for slice_i in slices), axis=t_axis)

    def add(self, array, filename=None):
        new = add_data(self.h5_file, self.name, array, filename)
        return new

    def subset(self, indices):
        if isinstance(indices, Metadata):
            indices = indices.all_indices
        return Subset(DataObject(self), indices)

    def __repr__(self):
        return "<FeatureProxy: '%s/%s'>" % (self.h5_file, self.name)


class Database(object):
    def __init__(self, h5_file):
        self.h5_file = h5_file
        self.info = self._get_dataframe("/info")
        self.metadata = Metadata(self._get_dataframe("/metadata"))
        with h5py.File(h5_file, "r") as f:
            # add found features as self.feature_name = FeatureProxy(self.h5_file, feature_name)
            self._register_features(self.features)
            self._register_dataframes(self.dataframes)

    @property
    def features(self):
        names = self.info.iloc[:, 2:].T.index.get_level_values(0)
        return list(set(names))

    @property
    def dataframes(self):
        keys = set()

        def func(k, v):
            if "pandas_type" in v.attrs.keys() and k.split("/")[0] not in ("layouts", "info", "metadata"):
                keys.add(k)
            return None

        self.visit(func)
        return list(keys)

    def visit(self, func=print):
        with h5py.File(self.h5_file, "r") as f:
            f.visititems(func)

    def _get_dataframe(self, key):
        try:
            return pd.read_hdf(self.h5_file, key=key)
        except KeyError:
            return pd.DataFrame()

    def save_dataframe(self, key, df):
        with h5py.File(self.h5_file, "r+") as f:
            if key in f:
                f.pop(key)
        df.to_hdf(self.h5_file, key=key, mode="r+")
        return self._get_dataframe(key)

    def layout_for(self, feature):
        with h5py.File(self.h5_file, "r") as f:
            if "layouts" not in f.keys():
                return pd.DataFrame()
        return self._get_dataframe("layouts/" + feature)

    def _register_features(self, names):
        for name in names:
            setattr(self, name, FeatureProxy(self.h5_file, name))
        return None

    def _register_dataframes(self, names):
        for name in names:
            setattr(self, name, self._get_dataframe(name))
        return None

    def __repr__(self):
        return "<Database: '%s'>" % os.path.split(self.h5_file)[-1]


def add_feature(h5_file, feature_name, array):
    # TODO !
    pass


def add_metadata(h5_file, start, stop, duration, ds_name, filename=None):
    meta = pd.read_hdf(h5_file, "metadata")
    layout = pd.read_hdf(h5_file, "layouts/" + ds_name)
    info = pd.read_hdf(h5_file, "info")
    new = Metadata.from_start_stop([start], [stop], [duration])
    filename = datetime.now() if filename is None else filename
    new["name"] = filename
    new_meta = pd.concat((meta, new), axis=0, ignore_index=True)
    new_layout = pd.concat((layout, new), axis=0, ignore_index=True)
    new_info = info.iloc[info.index.max()]
    new_info.loc[:] = ("added", filename,
                       new_info[(ds_name, "dtype")],
                       (new.duration.item(), *new_info[(ds_name, "shape")][1:]),
                       "xMb")
    new_info = info.append(new_info, ignore_index=True)
    new_meta.to_hdf(h5_file, "metadata")
    new_layout.to_hdf(h5_file, "layouts/" + ds_name)
    new_info.to_hdf(h5_file, "info")
    return new_info


def add_data(h5_file, ds_name, array, filename):
    N = array.shape[0]
    with h5py.File(h5_file, "r+") as f:
        if f[ds_name].shape[1:] != array.shape[1:]:
            raise ValueError(
                ("expected all but the first dimension of `array` to match %s. " % str(f[ds_name].shape[1:])) +
                ("Got %s" % str(array.shape[1:])))
        M = f[ds_name].shape[0]
        f[ds_name].resize(M + N, axis=0)
        f[ds_name][-N:] = array
        rv = f[ds_name][-N:]
    add_metadata(h5_file, M, N + M, N, ds_name, filename)
    return rv


def upload_database(db, api_token, project_name, experiment_name):
    session = Session.with_default_backend(api_token=api_token)
    data_project = session.get_project(project_name)
    feature = [name for name in db.features if "label" not in name][0]
    feat_prox = getattr(db, feature)
    params = {"name": experiment_name,
              "feature_name": feature,
              "shape": feat_prox.shape,
              "files": len(db.metadata)}
    params.update(feat_prox.attrs)
    exp = data_project.create_experiment(name=experiment_name,
                                         params=params)
    exp.log_artifact(db.h5_file)
    return exp.stop()


def download_database(api_token, full_exp_path, database_name, destination="./"):
    session = Session.with_default_backend(api_token=api_token)
    namespace, project, exp_id = full_exp_path.split("/")
    project_name = namespace + "/" + project
    data_project = session.get_project(project_name)
    exp = data_project.get_experiments(id=exp_id)[0]
    exp.download_artifact(database_name, destination)
    return Database(os.path.join(destination, database_name))



#################### Maintained by Hatch ####################
# This file is auto-generated by hatch. If you'd like to customize this file
# please add your changes near the bottom marked for 'USER OVERRIDES'.
# EVERYTHING ELSE WILL BE OVERWRITTEN by hatch.
#############################################################
from io import open

from setuptools import setup, find_packages
import os

with open('mimikit/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break
    else:
        version = '0.0.1'

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), "r", encoding="utf-8") as f:
    REQUIRES = [ln.strip() for ln in f.readlines() if ln.strip()]

PACKAGES = find_packages(exclude=('tests', 'tests.*'))

kwargs = {
    'name': 'mimikit',
    'version': version,
    'description': 'Python module for generating audio with neural networks',
    'long_description': readme,
    "long_description_content_type": "text/markdown",
    'author': 'Antoine Daurat',
    'author_email': 'ktonalberlin@gmail.com',
    'url': 'https://github.com/k-tonal/mmk',
    'download_url': 'https://github.com/k-tonal/mmk',
    # 'license': 'GNU General Public License v3 (GPLv3)',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        "Intended Audience :: Science/Research",
        "Intended Audience :: Other Audience",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Topic :: Multimedia :: Sound/Audio :: Sound Synthesis",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    "keywords": "audio music sound deep-learning",
    'python_requires': '>=3.6',
    'install_requires': REQUIRES,
    'tests_require': ['coverage', 'pytest'],
    'packages': PACKAGES,
    "entry_points": {
        'console_scripts': [
            'freqnet-db=mimikit.freqnet.freqnet_db:main'
        ]}

}

setup(**kwargs)



import librosa
from .metadata import Metadata

N_FFT = 2048
HOP_LENGTH = 512
SR = 22050
MU = 255


def stft(file, n_fft=N_FFT, hop_length=HOP_LENGTH, sr=SR):
    y, sr = librosa.load(file, sr=sr)
    fft = librosa.stft(y, n_fft=n_fft, hop_length=hop_length)
    # returns the feature and its attributes
    return fft, dict(n_fft=n_fft, hop_length=hop_length, sr=sr)


def file_to_fft(abs_path, n_fft=N_FFT, hop_length=HOP_LENGTH, sr=SR):
    fft, params = stft(abs_path, n_fft, hop_length, sr)
    fft = abs(fft)
    metadata = Metadata.from_duration([fft.shape[1]])
    params.update(dict(time_axis=0))
    return dict(fft=(params, fft.T), metadata=({}, metadata))


def mu_law_compress(file, mu=MU, sr=SR):
    y, sr = librosa.load(file, sr=sr)
    y = librosa.util.normalize(y)
    qx = librosa.mu_compress(y, mu, quantize=True)
    qx = qx + (MU + 1) // 2
    return qx, dict(mu=mu, sr=sr)


def file_to_qx(abs_path, mu=MU, sr=SR):
    qx, params = mu_law_compress(abs_path, mu, sr)
    metadata = Metadata.from_duration([qx.shape[0]])
    return dict(qx=(params, qx.reshape(-1, 1)), metadata=({}, metadata))


default_extract_func = file_to_fft



import torch.nn as nn
import numpy as np

from .modules import GatedLinearInput, AbsLinearOutput, mean_L1_prop
from .freq_layer import FreqLayer
from .base import FreqNetModel


class FreqNet(FreqNetModel):
    LAYER_KWARGS = ["groups", "strict", "accum_outputs", "concat_outputs",
                    "pad_input", "learn_padding", "with_skip_conv", "with_residual_conv"]

    def __init__(self,
                 loss_fn=mean_L1_prop,
                 model_dim=512,
                 groups=1,
                 n_layers=(2,),
                 strict=False,
                 accum_outputs=0,
                 concat_outputs=0,
                 pad_input=0,
                 learn_padding=False,
                 with_skip_conv=True,
                 with_residual_conv=True,
                 **data_optim_kwargs):
        super(FreqNet, self).__init__(**data_optim_kwargs)
        self._loss_fn = loss_fn
        self.model_dim = model_dim
        self.groups = groups
        self.n_layers = n_layers
        self.strict = strict
        self.accum_outputs = accum_outputs
        self.concat_outputs = concat_outputs
        self.pad_input = pad_input
        self.learn_padding = learn_padding
        self.with_skip_conv = with_skip_conv
        self.with_residual_conv = with_residual_conv

        # Input Encoder
        self.inpt = GatedLinearInput(self.input_dim, self.model_dim)

        # Auto-regressive Part
        layer_kwargs = {attr: getattr(self, attr) for attr in self.LAYER_KWARGS}
        # for simplicity we keep all the layers in a flat list
        self.layers = nn.ModuleList([
            FreqLayer(layer_index=i, input_dim=model_dim, layer_dim=model_dim, **layer_kwargs)
            for n_layers in self.n_layers for i in range(n_layers)
        ])

        # Output Decoder
        self.outpt = AbsLinearOutput(self.model_dim, self.input_dim)

        self.save_hyperparameters()

    def forward(self, x):
        """
        """
        x = self.inpt(x)
        skips = None
        for layer in self.layers:
            x, skips = layer(x, skips)
        x = self.outpt(skips)
        return x

    def loss_fn(self, predictions, targets):
        return self._loss_fn(predictions, targets)

    def all_rel_shifts(self):
        """sequence of shifts from one layer to the next"""
        return tuple(layer.rel_shift() for layer in self.layers)

    def shift(self):
        """total shift of the network"""
        if self.pad_input == 1 or self.concat_outputs == 1:
            return 1
        return sum(self.all_rel_shifts()) + int(not self.strict)

    def all_shifts(self):
        """the accumulated shift at each layer"""
        return tuple(np.cumsum(self.all_rel_shifts()) + int(not self.strict))

    def receptive_field(self):
        block_rf = []
        for i, layer in enumerate(self.layers[:-1]):
            if self.layers[i + 1].layer_index == 0:
                block_rf += [layer.receptive_field() - 1]
        block_rf += [self.layers[-1].receptive_field()]
        return sum(block_rf)

    def output_length(self, input_length):
        return self.all_output_lengths(input_length)[-1]

    def all_output_lengths(self, input_length):
        out_length = input_length
        lengths = []
        for layer in self.layers:
            out_length = layer.output_length(out_length)
            lengths += [out_length]
        return tuple(lengths)

    def targets_shifts_and_lengths(self, input_length):
        return [(self.shift(), self.output_length(input_length))]

    def generation_slices(self):
        # TODO! Add logic for concat/strict parameters
        return slice(-self.receptive_field(), None), slice(None, 1)



import torch
import torch.nn as nn
import numpy as np
import math

from .modules import GatedConv, LearnablePad1d


def accum(x, y, shift=1):
    if shift == 0:
        if x.size(-1) == y.size(-1):
            return x + y
        else:
            raise ValueError("zero shift with size: %i and %i" % (x.size(-1), y.size(-1)))
    side = np.sign(shift)
    aligned_x = slice(*((shift, None)[::side]))
    n_aligned = x.size(-1) - abs(shift)
    aligned_y = slice(*((None, n_aligned * side)[::side]))
    compl_y = slice(*((n_aligned * side, None)[::side]))
    aligned = x[:, :, aligned_x] + y[:, :, aligned_y]
    rest = y[:, :, compl_y]
    return torch.cat((aligned, rest)[::side], dim=-1)


def concat(x, y, shift=1):
    if shift == 0:
        if x.size(-1) != y.size(-1):
            return concat(x, y, - x.size(-1) - y.size(-1))
        return y
    side = np.sign(shift)
    compl_x = slice(*((None, shift)[::side]))
    return torch.cat((x[:, :, compl_x], y)[::side], dim=-1)


class FreqLayer(nn.Module):
    kernel_size = 2
    stride = 1
    bias = True
    sides = {1: 1, -1: -1, 0: 0, "left": 1, "right": -1}

    def __init__(self,
                 layer_index,
                 input_dim=512,
                 layer_dim=512,
                 groups=1,
                 strict=False,
                 accum_outputs=0,
                 concat_outputs=0,
                 pad_input=0,
                 learn_padding=False,
                 with_skip_conv=False,
                 with_residual_conv=False,
                 ):
        super(FreqLayer, self).__init__()
        self.layer_index = layer_index
        self.input_dim = input_dim
        self.layer_dim = layer_dim
        self.groups = groups
        self.strict = strict
        if pad_input != 0 and concat_outputs != 0:
            raise ValueError("pad_input and concat_outputs can not be both non zero")
        # accum and concat have opposite signs :
        # accum left is a shift from the end (negative), whereas concat left is a shift from the beginning (positive)
        self.accum_outputs = - self.sides.get(accum_outputs, 0)
        self.concat_outputs = self.sides.get(concat_outputs, 0)
        self.pad_input = self.sides.get(pad_input, 0)
        self.learn_padding = learn_padding
        self.with_skip_conv = with_skip_conv
        self.with_residual_conv = with_residual_conv

        self.pad = LearnablePad1d(self.input_dim, self.padding, self.learn_padding)

        convs_kwargs = dict(dilation=self.dilation, stride=self.stride,
                            groups=self.groups, bias=self.bias)
        self.gate = GatedConv(self.input_dim, self.layer_dim, self.kernel_size, **convs_kwargs)

        self.skips = nn.Conv1d(self.layer_dim, self.layer_dim, kernel_size=1, **convs_kwargs) \
            if self.with_skip_conv else None
        self.residuals = nn.Conv1d(self.layer_dim, self.input_dim, kernel_size=1, **convs_kwargs) \
            if self.with_residual_conv else None

    def forward(self, x, skip=None):
        x = self.pad(x)

        y = self.gate(x)
        h = self.skips(y) if self.with_skip_conv else None
        y = self.residuals(y) if self.with_residual_conv else y

        accum_outputs, concat_outputs = self.accum_outputs, self.concat_outputs
        shift = self.rel_shift()

        if accum_outputs:
            y = accum(x, y, shift * accum_outputs)

        if concat_outputs:
            y = concat(x, y, shift * concat_outputs)

        if self.with_skip_conv:
            if skip is None:
                skip = torch.zeros_like(h).to(h)
            if accum_outputs:
                # we do it on the same side as the outputs
                skip = accum(skip, h, shift * accum_outputs)
            else:
                # we accum on the right anyway, otherwise skips don't make any sense,
                # since they aren't inputs to anything...
                skip = accum(skip, h, shift * 1)
            if concat_outputs:
                skip = concat(skip, h, shift * concat_outputs)
        else:
            # still need to output a tensor
            skip = y

        return y, skip

    @property
    def dilation(self):
        return 2 ** self.layer_index

    @property
    def padding(self):
        """
        signed amount of padding necessary to output as many time-steps as there were in the inputs.
        The sign of padding corresponds to the desired side (1: left, -1: right)
        """
        return self.pad_input * (self.kernel_size - 1) * self.dilation

    def receptive_field(self):
        """
        amount of inputs necessary for 1 output (this is independent of the shift!)
        """
        return 2 ** (self.layer_index + 1)

    def shift(self):
        """total shift at this layer wrt. to the beginning of its block"""
        return self.receptive_field() + self.layer_index * int(self.strict)

    def rel_shift(self):
        """relative shift at this layer wrt. to the previous layer"""
        return int(self.strict) + self.receptive_field() // 2

    def output_length(self, input_length):
        if abs(self.concat_outputs):
            # concats appends past steps which grow by 1 at each layer when strict=True
            return input_length + int(self.strict)
        if abs(self.pad_input):
            # no matter what, padding input gives the same output shape
            return input_length
        # output is gonna be less than input
        numerator = input_length - self.dilation * (self.kernel_size - 1) - 1
        denominator = self.stride
        return math.floor(1 + numerator / denominator)
